#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Example Airflow DAG for testing Google Dataflow.

:class:`~airflow.providers.google.cloud.operators.dataflow.DataflowTemplatedJobStartOperator` operator.
"""

from __future__ import annotations

import os
from datetime import datetime
from pathlib import Path

from airflow.models.baseoperator import chain
from airflow.models.dag import DAG
from airflow.providers.google.cloud.operators.dataflow import (
    DataflowStartFlexTemplateOperator,
    DataflowTemplatedJobStartOperator,
)
from airflow.providers.google.cloud.operators.gcs import GCSCreateBucketOperator, GCSDeleteBucketOperator
from airflow.providers.google.cloud.transfers.local_to_gcs import LocalFilesystemToGCSOperator

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

from system.google import DEFAULT_GCP_SYSTEM_TEST_PROJECT_ID

ENV_ID = os.environ.get("SYSTEM_TESTS_ENV_ID", "default")
PROJECT_ID = os.environ.get("SYSTEM_TESTS_GCP_PROJECT") or DEFAULT_GCP_SYSTEM_TEST_PROJECT_ID
DAG_ID = "dataflow_template"

BUCKET_NAME = f"bucket_{DAG_ID}_{ENV_ID}".replace("_", "-")

CSV_FILE_NAME = "input.csv"
AVRO_FILE_NAME = "output.avro"
AVRO_SCHEMA = "schema.json"
GCS_TMP = f"gs://{BUCKET_NAME}/temp/"
GCS_STAGING = f"gs://{BUCKET_NAME}/staging/"
GCS_OUTPUT = f"gs://{BUCKET_NAME}/output"
PYTHON_FILE_LOCAL_PATH = str(Path(__file__).parent / "resources" / CSV_FILE_NAME)
SCHEMA_LOCAL_PATH = str(Path(__file__).parent / "resources" / AVRO_SCHEMA)
LOCATION = "europe-west3"

default_args = {
    "dataflow_default_options": {
        "tempLocation": GCS_TMP,
        "stagingLocation": GCS_STAGING,
    }
}
BODY = {
    "launchParameter": {
        "jobName": "test-flex-template",
        "parameters": {
            "inputFileSpec": f"gs://{BUCKET_NAME}/{CSV_FILE_NAME}",
            "outputBucket": f"gs://{BUCKET_NAME}/output/{AVRO_FILE_NAME}",
            "outputFileFormat": "avro",
            "inputFileFormat": "csv",
            "schema": f"gs://{BUCKET_NAME}/{AVRO_SCHEMA}",
        },
        "environment": {},
        "containerSpecGcsPath": "gs://dataflow-templates/latest/flex/File_Format_Conversion",
    },
}

with DAG(
    DAG_ID,
    default_args=default_args,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    catchup=False,
    tags=["example", "dataflow"],
) as dag:
    create_bucket = GCSCreateBucketOperator(task_id="create_bucket", bucket_name=BUCKET_NAME)

    upload_file = LocalFilesystemToGCSOperator(
        task_id="upload_file_to_bucket",
        src=PYTHON_FILE_LOCAL_PATH,
        dst=CSV_FILE_NAME,
        bucket=BUCKET_NAME,
    )

    upload_schema = LocalFilesystemToGCSOperator(
        task_id="upload_schema_to_bucket",
        src=SCHEMA_LOCAL_PATH,
        dst=AVRO_SCHEMA,
        bucket=BUCKET_NAME,
    )

    # [START howto_operator_start_template_job]
    start_template_job = DataflowTemplatedJobStartOperator(
        task_id="start_template_job",
        project_id=PROJECT_ID,
        template="gs://dataflow-templates/latest/Word_Count",
        parameters={"inputFile": f"gs://{BUCKET_NAME}/{CSV_FILE_NAME}", "output": GCS_OUTPUT},
        location=LOCATION,
        wait_until_finished=True,
    )
    # [END howto_operator_start_template_job]

    # [START howto_operator_start_flex_template_job]
    start_flex_template_job = DataflowStartFlexTemplateOperator(
        task_id="start_flex_template_job",
        project_id=PROJECT_ID,
        body=BODY,
        location=LOCATION,
        append_job_name=False,
        wait_until_finished=True,
    )
    # [END howto_operator_start_flex_template_job]

    # [START howto_operator_start_template_job_deferrable]
    start_template_job_deferrable = DataflowTemplatedJobStartOperator(
        task_id="start_template_job_deferrable",
        project_id=PROJECT_ID,
        template="gs://dataflow-templates/latest/Word_Count",
        parameters={"inputFile": f"gs://{BUCKET_NAME}/{CSV_FILE_NAME}", "output": GCS_OUTPUT},
        location=LOCATION,
        deferrable=True,
    )
    # [END howto_operator_start_template_job_deferrable]

    # [START howto_operator_start_flex_template_job_deferrable]
    start_flex_template_job_deferrable = DataflowStartFlexTemplateOperator(
        task_id="start_flex_template_job_deferrable",
        project_id=PROJECT_ID,
        body=BODY,
        location=LOCATION,
        append_job_name=False,
        deferrable=True,
    )
    # [END howto_operator_start_flex_template_job_deferrable]

    delete_bucket = GCSDeleteBucketOperator(
        task_id="delete_bucket", bucket_name=BUCKET_NAME, trigger_rule=TriggerRule.ALL_DONE
    )

    chain(
        create_bucket,
        upload_file,
        upload_schema,
        [start_template_job, start_flex_template_job],
        [start_template_job_deferrable, start_flex_template_job_deferrable],
        delete_bucket,
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "teardown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()


from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import json
import os
import subprocess

from airflow import settings
from airflow.exceptions import AirflowException
from airflow.models import Connection

from tests_common.test_utils import AIRFLOW_MAIN_FOLDER
from tests_common.test_utils.logging_command_executor import CommandExecutor

# Please keep these variables in alphabetical order.

GCP_AI_KEY = "gcp_ai.json"
GCP_BIGQUERY_KEY = "gcp_bigquery.json"
GCP_BIGTABLE_KEY = "gcp_bigtable.json"
GCP_CLOUD_BUILD_KEY = "gcp_cloud_build.json"
GCP_CLOUD_COMPOSER = "gcp_cloud_composer.json"
GCP_CLOUDSQL_KEY = "gcp_cloudsql.json"
GCP_COMPUTE_KEY = "gcp_compute.json"
GCP_COMPUTE_SSH_KEY = "gcp_compute_ssh.json"
GCP_DATACATALOG_KEY = "gcp_datacatalog.json"
GCP_DATAFLOW_KEY = "gcp_dataflow.json"
GCP_DATAFUSION_KEY = "gcp_datafusion.json"
GCP_DATAPROC_KEY = "gcp_dataproc.json"
GCP_DATASTORE_KEY = "gcp_datastore.json"
GCP_DLP_KEY = "gcp_dlp.json"
GCP_FUNCTION_KEY = "gcp_function.json"
GCP_GCS_KEY = "gcp_gcs.json"
GCP_GCS_TRANSFER_KEY = "gcp_gcs_transfer.json"
GCP_GKE_KEY = "gcp_gke.json"
GCP_KMS_KEY = "gcp_kms.json"
GCP_MEMORYSTORE = "gcp_memorystore.json"
GCP_PUBSUB_KEY = "gcp_pubsub.json"
GCP_SECRET_MANAGER_KEY = "gcp_secret_manager.json"
GCP_SPANNER_KEY = "gcp_spanner.json"
GCP_STACKDRIVER = "gcp_stackdriver.json"
GCP_TASKS_KEY = "gcp_tasks.json"
GCP_VERTEX_AI_KEY = "gcp_vertex_ai.json"
GCP_WORKFLOWS_KEY = "gcp_workflows.json"
GMP_KEY = "gmp.json"
G_FIREBASE_KEY = "g_firebase.json"
GCP_AWS_KEY = "gcp_aws.json"

KEYPATH_EXTRA = "extra__google_cloud_platform__key_path"
KEYFILE_DICT_EXTRA = "extra__google_cloud_platform__keyfile_dict"
SCOPE_EXTRA = "extra__google_cloud_platform__scope"
PROJECT_EXTRA = "extra__google_cloud_platform__project"


class GcpAuthenticator(CommandExecutor):
    """
    Initialises the authenticator.

    :param gcp_key: name of the key to use for authentication (see GCP_*_KEY values)
    :param project_extra: optional extra project parameter passed to google cloud
           connection
    """

    original_account: str | None = None

    def __init__(self, gcp_key: str, project_extra: str | None = None):
        super().__init__()
        self.gcp_key = gcp_key
        self.project_extra = project_extra
        self.project_id = self.get_project_id()
        self.full_key_path = None
        self._set_key_path()

    @staticmethod
    def get_project_id():
        return os.environ.get("GCP_PROJECT_ID")

    def set_key_path_in_airflow_connection(self):
        """
        Set key path in 'google_cloud_default' connection to point to the full
        key path
        :return: None
        """
        with settings.Session() as session:
            conn = session.query(Connection).filter(Connection.conn_id == "google_cloud_default")[0]
            extras = conn.extra_dejson
            extras[KEYPATH_EXTRA] = self.full_key_path
            if extras.get(KEYFILE_DICT_EXTRA):
                del extras[KEYFILE_DICT_EXTRA]
            extras[SCOPE_EXTRA] = "https://www.googleapis.com/auth/cloud-platform"
            extras[PROJECT_EXTRA] = self.project_extra if self.project_extra else self.project_id
            conn.extra = json.dumps(extras)

    def set_dictionary_in_airflow_connection(self):
        """
        Set dictionary in 'google_cloud_default' connection to contain content
        of the json service account file.
        :return: None
        """
        with settings.Session() as session:
            conn = session.query(Connection).filter(Connection.conn_id == "google_cloud_default")[0]
            extras = conn.extra_dejson
            with open(self.full_key_path) as path_file:
                content = json.load(path_file)
            extras[KEYFILE_DICT_EXTRA] = json.dumps(content)
            if extras.get(KEYPATH_EXTRA):
                del extras[KEYPATH_EXTRA]
            extras[SCOPE_EXTRA] = "https://www.googleapis.com/auth/cloud-platform"
            extras[PROJECT_EXTRA] = self.project_extra
            conn.extra = json.dumps(extras)

    def _set_key_path(self):
        """
        Sets full key path - if GCP_CONFIG_DIR points to absolute
            directory, it tries to find the key in this directory. Otherwise it assumes
            that Airflow is running from the directory where configuration is checked
            out next to airflow directory in config directory
            it tries to find the key folder in the workspace's config
            directory.
        :param : name of the key file to find.
        """
        if "GCP_CONFIG_DIR" in os.environ:
            gcp_config_dir = os.environ["GCP_CONFIG_DIR"]
        else:
            gcp_config_dir = os.path.join(AIRFLOW_MAIN_FOLDER, os.pardir, "config")
        if not os.path.isdir(gcp_config_dir):
            self.log.info("The %s is not a directory", gcp_config_dir)
        key_dir = os.path.join(gcp_config_dir, "keys")
        if not os.path.isdir(key_dir):
            self.log.error("The %s is not a directory", key_dir)
            return
        key_path = os.path.join(key_dir, self.gcp_key)
        if not os.path.isfile(key_path):
            self.log.error("The %s file is missing", key_path)
        self.full_key_path = key_path

    def _validate_key_set(self):
        if self.full_key_path is None:
            raise AirflowException("The gcp_key is not set!")
        if not os.path.isfile(self.full_key_path):
            raise AirflowException(
                f"The key {self.gcp_key} could not be found. Please copy it to the {self.full_key_path} path."
            )

    def gcp_authenticate(self):
        """
        Authenticate with service account specified via key name.
        """
        self._validate_key_set()
        self.log.info("Setting the Google Cloud key to %s", self.full_key_path)
        # Checking if we can authenticate using service account credentials provided
        self.execute_cmd(
            [
                "gcloud",
                "auth",
                "activate-service-account",
                f"--key-file={self.full_key_path}",
                f"--project={self.project_id}",
            ]
        )
        self.set_key_path_in_airflow_connection()

    def gcp_revoke_authentication(self):
        """
        Change default authentication to none - which is not existing one.
        """
        self._validate_key_set()
        self.log.info("Revoking authentication - setting it to none")
        self.execute_cmd(["gcloud", "config", "get-value", "account", f"--project={self.project_id}"])
        self.execute_cmd(["gcloud", "config", "set", "account", "none", f"--project={self.project_id}"])

    def gcp_store_authentication(self):
        """
        Store authentication as it was originally so it can be restored and revoke
        authentication.
        """
        self._validate_key_set()
        if not GcpAuthenticator.original_account:
            GcpAuthenticator.original_account = self.check_output(
                ["gcloud", "config", "get-value", "account", f"--project={self.project_id}"]
            ).decode("utf-8")
            self.log.info("Storing account: to restore it later %s", GcpAuthenticator.original_account)

    def gcp_restore_authentication(self):
        """
        Restore authentication to the original one.
        """
        self._validate_key_set()
        if GcpAuthenticator.original_account:
            self.log.info("Restoring original account stored: %s", GcpAuthenticator.original_account)
            subprocess.call(
                [
                    "gcloud",
                    "config",
                    "set",
                    "account",
                    GcpAuthenticator.original_account,
                    f"--project={self.project_id}",
                ]
            )
        else:
            self.log.info("Not restoring the original Google Cloud account: it is not set")



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module contains Google Datastore operators."""

from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING, Any

from airflow.exceptions import AirflowException
from airflow.providers.google.cloud.hooks.datastore import DatastoreHook
from airflow.providers.google.cloud.hooks.gcs import GCSHook
from airflow.providers.google.cloud.links.datastore import (
    CloudDatastoreEntitiesLink,
    CloudDatastoreImportExportLink,
)
from airflow.providers.google.cloud.operators.cloud_base import GoogleCloudBaseOperator
from airflow.providers.google.common.hooks.base_google import PROVIDE_PROJECT_ID
from airflow.providers.google.common.links.storage import StorageLink

if TYPE_CHECKING:
    from airflow.utils.context import Context


class CloudDatastoreExportEntitiesOperator(GoogleCloudBaseOperator):
    """
    Export entities from Google Cloud Datastore to Cloud Storage.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:CloudDatastoreExportEntitiesOperator`

    .. seealso::
        https://cloud.google.com/datastore/docs/export-import-entities

    :param bucket: name of the cloud storage bucket to back up data
    :param namespace: optional namespace path in the specified Cloud Storage bucket
        to back up data. If this namespace does not exist in GCS, it will be created.
    :param datastore_conn_id: the name of the Datastore connection id to use
    :param cloud_storage_conn_id: the name of the cloud storage connection id to
        force-write backup
    :param entity_filter: description of what data from the project is included in the
        export, refer to
        https://cloud.google.com/datastore/docs/reference/rest/Shared.Types/EntityFilter
    :param labels: client-assigned labels for cloud storage
    :param polling_interval_in_seconds: number of seconds to wait before polling for
        execution status again
    :param overwrite_existing: if the storage bucket + namespace is not empty, it will be
        emptied prior to exports. This enables overwriting existing backups.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = (
        "bucket",
        "namespace",
        "entity_filter",
        "labels",
        "impersonation_chain",
    )
    operator_extra_links = (StorageLink(),)

    def __init__(
        self,
        *,
        bucket: str,
        namespace: str | None = None,
        datastore_conn_id: str = "google_cloud_default",
        cloud_storage_conn_id: str = "google_cloud_default",
        entity_filter: dict | None = None,
        labels: dict | None = None,
        polling_interval_in_seconds: int = 10,
        overwrite_existing: bool = False,
        project_id: str = PROVIDE_PROJECT_ID,
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.datastore_conn_id = datastore_conn_id
        self.cloud_storage_conn_id = cloud_storage_conn_id
        self.bucket = bucket
        self.namespace = namespace
        self.entity_filter = entity_filter
        self.labels = labels
        self.polling_interval_in_seconds = polling_interval_in_seconds
        self.overwrite_existing = overwrite_existing
        self.project_id = project_id
        self.impersonation_chain = impersonation_chain

    def execute(self, context: Context) -> dict:
        self.log.info("Exporting data to Cloud Storage bucket %s", self.bucket)

        if self.overwrite_existing and self.namespace:
            gcs_hook = GCSHook(self.cloud_storage_conn_id, impersonation_chain=self.impersonation_chain)
            objects = gcs_hook.list(self.bucket, prefix=self.namespace)
            for obj in objects:
                gcs_hook.delete(self.bucket, obj)

        ds_hook = DatastoreHook(
            gcp_conn_id=self.datastore_conn_id,
            impersonation_chain=self.impersonation_chain,
        )
        result = ds_hook.export_to_storage_bucket(
            bucket=self.bucket,
            namespace=self.namespace,
            entity_filter=self.entity_filter,
            labels=self.labels,
            project_id=self.project_id,
        )
        operation_name = result["name"]
        result = ds_hook.poll_operation_until_done(operation_name, self.polling_interval_in_seconds)

        state = result["metadata"]["common"]["state"]
        if state != "SUCCESSFUL":
            raise AirflowException(f"Operation failed: result={result}")
        StorageLink.persist(
            context=context,
            uri=f"{self.bucket}/{result['response']['outputUrl'].split('/')[3]}",
            project_id=self.project_id or ds_hook.project_id,
        )
        return result


class CloudDatastoreImportEntitiesOperator(GoogleCloudBaseOperator):
    """
    Import entities from Cloud Storage to Google Cloud Datastore.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:CloudDatastoreImportEntitiesOperator`

    .. seealso::
        https://cloud.google.com/datastore/docs/export-import-entities

    :param bucket: container in Cloud Storage to store data
    :param file: path of the backup metadata file in the specified Cloud Storage bucket.
        It should have the extension .overall_export_metadata
    :param namespace: optional namespace of the backup metadata file in
        the specified Cloud Storage bucket.
    :param entity_filter: description of what data from the project is included in
        the export, refer to
        https://cloud.google.com/datastore/docs/reference/rest/Shared.Types/EntityFilter
    :param labels: client-assigned labels for cloud storage
    :param datastore_conn_id: the name of the connection id to use
    :param polling_interval_in_seconds: number of seconds to wait before polling for
        execution status again
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = (
        "bucket",
        "file",
        "namespace",
        "entity_filter",
        "labels",
        "impersonation_chain",
    )
    operator_extra_links = (CloudDatastoreImportExportLink(),)

    def __init__(
        self,
        *,
        bucket: str,
        file: str,
        namespace: str | None = None,
        entity_filter: dict | None = None,
        labels: dict | None = None,
        datastore_conn_id: str = "google_cloud_default",
        polling_interval_in_seconds: float = 10,
        project_id: str = PROVIDE_PROJECT_ID,
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.datastore_conn_id = datastore_conn_id
        self.bucket = bucket
        self.file = file
        self.namespace = namespace
        self.entity_filter = entity_filter
        self.labels = labels
        self.polling_interval_in_seconds = polling_interval_in_seconds
        self.project_id = project_id
        self.impersonation_chain = impersonation_chain

    @property
    def extra_links_params(self) -> dict[str, Any]:
        return {
            "project_id": self.project_id,
        }

    def execute(self, context: Context):
        self.log.info("Importing data from Cloud Storage bucket %s", self.bucket)
        ds_hook = DatastoreHook(
            self.datastore_conn_id,
            impersonation_chain=self.impersonation_chain,
        )
        result = ds_hook.import_from_storage_bucket(
            bucket=self.bucket,
            file=self.file,
            namespace=self.namespace,
            entity_filter=self.entity_filter,
            labels=self.labels,
            project_id=self.project_id,
        )
        operation_name = result["name"]
        result = ds_hook.poll_operation_until_done(operation_name, self.polling_interval_in_seconds)

        state = result["metadata"]["common"]["state"]
        if state != "SUCCESSFUL":
            raise AirflowException(f"Operation failed: result={result}")
        CloudDatastoreImportExportLink.persist(context=context)
        return result


class CloudDatastoreAllocateIdsOperator(GoogleCloudBaseOperator):
    """
    Allocate IDs for incomplete keys. Return list of keys.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:CloudDatastoreAllocateIdsOperator`

    .. seealso::
        https://cloud.google.com/datastore/docs/reference/rest/v1/projects/allocateIds

    :param partial_keys: a list of partial keys.
    :param project_id: Google Cloud project ID against which to make the request.
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = (
        "partial_keys",
        "impersonation_chain",
    )
    operator_extra_links = (CloudDatastoreEntitiesLink(),)

    def __init__(
        self,
        *,
        partial_keys: list,
        project_id: str = PROVIDE_PROJECT_ID,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        self.partial_keys = partial_keys
        self.gcp_conn_id = gcp_conn_id
        self.project_id = project_id
        self.impersonation_chain = impersonation_chain

    @property
    def extra_links_params(self) -> dict[str, Any]:
        return {
            "project_id": self.project_id,
        }

    def execute(self, context: Context) -> list:
        hook = DatastoreHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )
        keys = hook.allocate_ids(
            partial_keys=self.partial_keys,
            project_id=self.project_id,
        )
        CloudDatastoreEntitiesLink.persist(context=context)
        return keys


class CloudDatastoreBeginTransactionOperator(GoogleCloudBaseOperator):
    """
    Begins a new transaction. Returns a transaction handle.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:CloudDatastoreBeginTransactionOperator`

    .. seealso::
        https://cloud.google.com/datastore/docs/reference/rest/v1/projects/beginTransaction

    :param transaction_options: Options for a new transaction.
    :param project_id: Google Cloud project ID against which to make the request.
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = (
        "transaction_options",
        "impersonation_chain",
    )

    def __init__(
        self,
        *,
        transaction_options: dict[str, Any],
        project_id: str = PROVIDE_PROJECT_ID,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        self.transaction_options = transaction_options
        self.gcp_conn_id = gcp_conn_id
        self.project_id = project_id
        self.impersonation_chain = impersonation_chain

    def execute(self, context: Context) -> str:
        hook = DatastoreHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )
        handle = hook.begin_transaction(
            transaction_options=self.transaction_options,
            project_id=self.project_id,
        )
        return handle


class CloudDatastoreCommitOperator(GoogleCloudBaseOperator):
    """
    Commit a transaction, optionally creating, deleting or modifying some entities.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:CloudDatastoreCommitOperator`

    .. seealso::
        https://cloud.google.com/datastore/docs/reference/rest/v1/projects/commit

    :param body: the body of the commit request.
    :param project_id: Google Cloud project ID against which to make the request.
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = (
        "body",
        "impersonation_chain",
    )
    operator_extra_links = (CloudDatastoreEntitiesLink(),)

    def __init__(
        self,
        *,
        body: dict[str, Any],
        project_id: str = PROVIDE_PROJECT_ID,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        self.body = body
        self.gcp_conn_id = gcp_conn_id
        self.project_id = project_id
        self.impersonation_chain = impersonation_chain

    @property
    def extra_links_params(self) -> dict[str, Any]:
        return {
            "project_id": self.project_id,
        }

    def execute(self, context: Context) -> dict:
        hook = DatastoreHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )
        response = hook.commit(
            body=self.body,
            project_id=self.project_id,
        )
        CloudDatastoreEntitiesLink.persist(context=context)
        return response


class CloudDatastoreRollbackOperator(GoogleCloudBaseOperator):
    """
    Roll back a transaction.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:CloudDatastoreRollbackOperator`

    .. seealso::
        https://cloud.google.com/datastore/docs/reference/rest/v1/projects/rollback

    :param transaction: the transaction to roll back.
    :param project_id: Google Cloud project ID against which to make the request.
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = (
        "transaction",
        "impersonation_chain",
    )

    def __init__(
        self,
        *,
        transaction: str,
        project_id: str = PROVIDE_PROJECT_ID,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        self.transaction = transaction
        self.gcp_conn_id = gcp_conn_id
        self.project_id = project_id
        self.impersonation_chain = impersonation_chain

    def execute(self, context: Context) -> None:
        hook = DatastoreHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )
        hook.rollback(
            transaction=self.transaction,
            project_id=self.project_id,
        )


class CloudDatastoreRunQueryOperator(GoogleCloudBaseOperator):
    """
    Run a query for entities. Returns the batch of query results.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:CloudDatastoreRunQueryOperator`

    .. seealso::
        https://cloud.google.com/datastore/docs/reference/rest/v1/projects/runQuery

    :param body: the body of the query request.
    :param project_id: Google Cloud project ID against which to make the request.
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = (
        "body",
        "impersonation_chain",
    )

    def __init__(
        self,
        *,
        body: dict[str, Any],
        project_id: str = PROVIDE_PROJECT_ID,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        self.body = body
        self.gcp_conn_id = gcp_conn_id
        self.project_id = project_id
        self.impersonation_chain = impersonation_chain

    def execute(self, context: Context) -> dict:
        hook = DatastoreHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )
        response = hook.run_query(
            body=self.body,
            project_id=self.project_id,
        )
        return response


class CloudDatastoreGetOperationOperator(GoogleCloudBaseOperator):
    """
    Gets the latest state of a long-running operation.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:CloudDatastoreGetOperationOperator`

    .. seealso::
        https://cloud.google.com/datastore/docs/reference/data/rest/v1/projects.operations/get

    :param name: the name of the operation resource.
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = (
        "name",
        "impersonation_chain",
    )

    def __init__(
        self,
        *,
        name: str,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        self.name = name
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

    def execute(self, context: Context):
        hook = DatastoreHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )
        op = hook.get_operation(name=self.name)
        return op


class CloudDatastoreDeleteOperationOperator(GoogleCloudBaseOperator):
    """
    Deletes the long-running operation.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:CloudDatastoreDeleteOperationOperator`

    .. seealso::
        https://cloud.google.com/datastore/docs/reference/data/rest/v1/projects.operations/delete

    :param name: the name of the operation resource.
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = (
        "name",
        "impersonation_chain",
    )

    def __init__(
        self,
        *,
        name: str,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        self.name = name
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

    def execute(self, context: Context) -> None:
        hook = DatastoreHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )
        hook.delete_operation(name=self.name)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock
from unittest.mock import AsyncMock

import pytest
from google.api_core.gapic_v1.method import DEFAULT
from google.cloud.orchestration.airflow.service_v1 import EnvironmentsAsyncClient

from airflow.providers.google.cloud.hooks.cloud_composer import CloudComposerAsyncHook, CloudComposerHook

TEST_GCP_REGION = "global"
TEST_GCP_PROJECT = "test-project"
TEST_GCP_CONN_ID = "test-gcp-conn-id"
TEST_ENVIRONMENT_ID = "testenvname"
TEST_ENVIRONMENT = {
    "name": TEST_ENVIRONMENT_ID,
    "config": {
        "node_count": 3,
        "software_config": {"image_version": "composer-1.17.7-airflow-2.1.4"},
    },
}
TEST_COMMAND = "dags"
TEST_SUBCOMMAND = "list"
TEST_PARAMETERS = ["--verbose", "-o", "json"]
TEST_EXECUTION_ID = "test-execution-id"
TEST_POD = "test-pod"
TEST_POD_NAMESPACE = "test-namespace"

TEST_UPDATE_MASK = {"paths": ["labels.label1"]}
TEST_UPDATED_ENVIRONMENT = {
    "labels": {
        "label1": "testing",
    }
}
TEST_RETRY = DEFAULT
TEST_TIMEOUT = None
TEST_METADATA = [("key", "value")]
TEST_PARENT = "test-parent"
TEST_NAME = "test-name"

BASE_STRING = "airflow.providers.google.common.hooks.base_google.{}"
COMPOSER_STRING = "airflow.providers.google.cloud.hooks.cloud_composer.{}"


def mock_init(*args, **kwargs):
    pass


class TestCloudComposerHook:
    def setup_method(self):
        with mock.patch(BASE_STRING.format("GoogleBaseHook.__init__"), new=mock_init):
            self.hook = CloudComposerHook(gcp_conn_id="test")

    @mock.patch(COMPOSER_STRING.format("CloudComposerHook.get_environment_client"))
    def test_create_environment(self, mock_client) -> None:
        self.hook.create_environment(
            project_id=TEST_GCP_PROJECT,
            region=TEST_GCP_REGION,
            environment=TEST_ENVIRONMENT,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        mock_client.assert_called_once()
        mock_client.return_value.create_environment.assert_called_once_with(
            request={
                "parent": self.hook.get_parent(TEST_GCP_PROJECT, TEST_GCP_REGION),
                "environment": TEST_ENVIRONMENT,
            },
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )

    @mock.patch(COMPOSER_STRING.format("CloudComposerHook.get_environment_client"))
    def test_delete_environment(self, mock_client) -> None:
        self.hook.delete_environment(
            project_id=TEST_GCP_PROJECT,
            region=TEST_GCP_REGION,
            environment_id=TEST_ENVIRONMENT_ID,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        mock_client.assert_called_once()
        mock_client.return_value.delete_environment.assert_called_once_with(
            request={
                "name": self.hook.get_environment_name(TEST_GCP_PROJECT, TEST_GCP_REGION, TEST_ENVIRONMENT_ID)
            },
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )

    @mock.patch(COMPOSER_STRING.format("CloudComposerHook.get_environment_client"))
    def test_get_environment(self, mock_client) -> None:
        self.hook.get_environment(
            project_id=TEST_GCP_PROJECT,
            region=TEST_GCP_REGION,
            environment_id=TEST_ENVIRONMENT_ID,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        mock_client.assert_called_once()
        mock_client.return_value.get_environment.assert_called_once_with(
            request={
                "name": self.hook.get_environment_name(TEST_GCP_PROJECT, TEST_GCP_REGION, TEST_ENVIRONMENT_ID)
            },
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )

    @mock.patch(COMPOSER_STRING.format("CloudComposerHook.get_environment_client"))
    def test_list_environments(self, mock_client) -> None:
        self.hook.list_environments(
            project_id=TEST_GCP_PROJECT,
            region=TEST_GCP_REGION,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        mock_client.assert_called_once()
        mock_client.return_value.list_environments.assert_called_once_with(
            request={
                "parent": self.hook.get_parent(TEST_GCP_PROJECT, TEST_GCP_REGION),
                "page_size": None,
                "page_token": None,
            },
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )

    @mock.patch(COMPOSER_STRING.format("CloudComposerHook.get_environment_client"))
    def test_update_environment(self, mock_client) -> None:
        self.hook.update_environment(
            project_id=TEST_GCP_PROJECT,
            region=TEST_GCP_REGION,
            environment_id=TEST_ENVIRONMENT_ID,
            environment=TEST_UPDATED_ENVIRONMENT,
            update_mask=TEST_UPDATE_MASK,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        mock_client.assert_called_once()
        mock_client.return_value.update_environment.assert_called_once_with(
            request={
                "name": self.hook.get_environment_name(
                    TEST_GCP_PROJECT, TEST_GCP_REGION, TEST_ENVIRONMENT_ID
                ),
                "environment": TEST_UPDATED_ENVIRONMENT,
                "update_mask": TEST_UPDATE_MASK,
            },
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )

    @mock.patch(COMPOSER_STRING.format("CloudComposerHook.get_image_versions_client"))
    def test_list_image_versions(self, mock_client) -> None:
        self.hook.list_image_versions(
            project_id=TEST_GCP_PROJECT,
            region=TEST_GCP_REGION,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        mock_client.assert_called_once()
        mock_client.return_value.list_image_versions.assert_called_once_with(
            request={
                "parent": self.hook.get_parent(TEST_GCP_PROJECT, TEST_GCP_REGION),
                "page_size": None,
                "page_token": None,
                "include_past_releases": False,
            },
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )

    @mock.patch(COMPOSER_STRING.format("CloudComposerHook.get_environment_client"))
    def test_execute_airflow_command(self, mock_client) -> None:
        self.hook.execute_airflow_command(
            project_id=TEST_GCP_PROJECT,
            region=TEST_GCP_REGION,
            environment_id=TEST_ENVIRONMENT_ID,
            command=TEST_COMMAND,
            subcommand=TEST_SUBCOMMAND,
            parameters=TEST_PARAMETERS,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        mock_client.assert_called_once()
        mock_client.return_value.execute_airflow_command.assert_called_once_with(
            request={
                "environment": self.hook.get_environment_name(
                    TEST_GCP_PROJECT, TEST_GCP_REGION, TEST_ENVIRONMENT_ID
                ),
                "command": TEST_COMMAND,
                "subcommand": TEST_SUBCOMMAND,
                "parameters": TEST_PARAMETERS,
            },
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )

    @mock.patch(COMPOSER_STRING.format("CloudComposerHook.get_environment_client"))
    def test_poll_airflow_command(self, mock_client) -> None:
        self.hook.poll_airflow_command(
            project_id=TEST_GCP_PROJECT,
            region=TEST_GCP_REGION,
            environment_id=TEST_ENVIRONMENT_ID,
            execution_id=TEST_EXECUTION_ID,
            pod=TEST_POD,
            pod_namespace=TEST_POD_NAMESPACE,
            next_line_number=1,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        mock_client.assert_called_once()
        mock_client.return_value.poll_airflow_command.assert_called_once_with(
            request={
                "environment": self.hook.get_environment_name(
                    TEST_GCP_PROJECT, TEST_GCP_REGION, TEST_ENVIRONMENT_ID
                ),
                "execution_id": TEST_EXECUTION_ID,
                "pod": TEST_POD,
                "pod_namespace": TEST_POD_NAMESPACE,
                "next_line_number": 1,
            },
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )


class TestCloudComposerAsyncHook:
    def setup_method(self, method):
        with mock.patch(BASE_STRING.format("GoogleBaseHook.__init__"), new=mock_init):
            self.hook = CloudComposerAsyncHook(gcp_conn_id="test")

    @pytest.mark.asyncio
    @mock.patch(COMPOSER_STRING.format("CloudComposerAsyncHook.get_environment_client"))
    async def test_create_environment(self, mock_client) -> None:
        mock_env_client = AsyncMock(EnvironmentsAsyncClient)
        mock_client.return_value = mock_env_client
        await self.hook.create_environment(
            project_id=TEST_GCP_PROJECT,
            region=TEST_GCP_REGION,
            environment=TEST_ENVIRONMENT,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        mock_client.assert_called_once()
        mock_client.return_value.create_environment.assert_called_once_with(
            request={
                "parent": self.hook.get_parent(TEST_GCP_PROJECT, TEST_GCP_REGION),
                "environment": TEST_ENVIRONMENT,
            },
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )

    @pytest.mark.asyncio
    @mock.patch(COMPOSER_STRING.format("CloudComposerAsyncHook.get_environment_client"))
    async def test_delete_environment(self, mock_client) -> None:
        mock_env_client = AsyncMock(EnvironmentsAsyncClient)
        mock_client.return_value = mock_env_client
        await self.hook.delete_environment(
            project_id=TEST_GCP_PROJECT,
            region=TEST_GCP_REGION,
            environment_id=TEST_ENVIRONMENT_ID,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        mock_client.assert_called_once()
        mock_client.return_value.delete_environment.assert_called_once_with(
            request={
                "name": self.hook.get_environment_name(TEST_GCP_PROJECT, TEST_GCP_REGION, TEST_ENVIRONMENT_ID)
            },
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )

    @pytest.mark.asyncio
    @mock.patch(COMPOSER_STRING.format("CloudComposerAsyncHook.get_environment_client"))
    async def test_update_environment(self, mock_client) -> None:
        mock_env_client = AsyncMock(EnvironmentsAsyncClient)
        mock_client.return_value = mock_env_client
        await self.hook.update_environment(
            project_id=TEST_GCP_PROJECT,
            region=TEST_GCP_REGION,
            environment_id=TEST_ENVIRONMENT_ID,
            environment=TEST_UPDATED_ENVIRONMENT,
            update_mask=TEST_UPDATE_MASK,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        mock_client.assert_called_once()
        mock_client.return_value.update_environment.assert_called_once_with(
            request={
                "name": self.hook.get_environment_name(
                    TEST_GCP_PROJECT, TEST_GCP_REGION, TEST_ENVIRONMENT_ID
                ),
                "environment": TEST_UPDATED_ENVIRONMENT,
                "update_mask": TEST_UPDATE_MASK,
            },
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )

    @pytest.mark.asyncio
    @mock.patch(COMPOSER_STRING.format("CloudComposerAsyncHook.get_environment_client"))
    async def test_execute_airflow_command(self, mock_client) -> None:
        mock_env_client = AsyncMock(EnvironmentsAsyncClient)
        mock_client.return_value = mock_env_client
        await self.hook.execute_airflow_command(
            project_id=TEST_GCP_PROJECT,
            region=TEST_GCP_REGION,
            environment_id=TEST_ENVIRONMENT_ID,
            command=TEST_COMMAND,
            subcommand=TEST_SUBCOMMAND,
            parameters=TEST_PARAMETERS,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        mock_client.assert_called_once()
        mock_client.return_value.execute_airflow_command.assert_called_once_with(
            request={
                "environment": self.hook.get_environment_name(
                    TEST_GCP_PROJECT, TEST_GCP_REGION, TEST_ENVIRONMENT_ID
                ),
                "command": TEST_COMMAND,
                "subcommand": TEST_SUBCOMMAND,
                "parameters": TEST_PARAMETERS,
            },
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )

    @pytest.mark.asyncio
    @mock.patch(COMPOSER_STRING.format("CloudComposerAsyncHook.get_environment_client"))
    async def test_poll_airflow_command(self, mock_client) -> None:
        mock_env_client = AsyncMock(EnvironmentsAsyncClient)
        mock_client.return_value = mock_env_client
        await self.hook.poll_airflow_command(
            project_id=TEST_GCP_PROJECT,
            region=TEST_GCP_REGION,
            environment_id=TEST_ENVIRONMENT_ID,
            execution_id=TEST_EXECUTION_ID,
            pod=TEST_POD,
            pod_namespace=TEST_POD_NAMESPACE,
            next_line_number=1,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        mock_client.assert_called_once()
        mock_client.return_value.poll_airflow_command.assert_called_once_with(
            request={
                "environment": self.hook.get_environment_name(
                    TEST_GCP_PROJECT, TEST_GCP_REGION, TEST_ENVIRONMENT_ID
                ),
                "execution_id": TEST_EXECUTION_ID,
                "pod": TEST_POD,
                "pod_namespace": TEST_POD_NAMESPACE,
                "next_line_number": 1,
            },
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import os
from unittest import mock

import pytest

from airflow.providers.google.cloud.transfers.mssql_to_gcs import MSSQLToGCSOperator
from airflow.providers.microsoft.mssql.hooks.mssql import MsSqlHook

AIRFLOW_CONN_MSSQL_DEFAULT = "mssql://sa:airflow123@mssql:1433/"

TASK_ID = "test-mssql-to-gcs"
TEST_TABLE_ID = "test_table_id"
MSSQL_CONN_ID = "mssql_default"
BUCKET = "gs://test"
JSON_FILENAME = "test_{}.ndjson"
GZIP = False

NDJSON_LINES = [
    b'{"Address": "1000 N West Street, Suite 1200", "City": "Wilmington, NC, USA", "FirstName": "Apache",'
    b' "LastName": "Airflow", "PersonID": 0}\n',
]


@pytest.mark.integration("mssql")
class TestMsSqlToGoogleCloudStorageOperator:
    def setup_method(self):
        os.environ["AIRFLOW_CONN_MSSQL_DEFAULT"] = AIRFLOW_CONN_MSSQL_DEFAULT
        hook = MsSqlHook()
        conn = hook.get_conn()
        hook.set_autocommit(conn, True)
        self.cursor = conn.cursor()
        self.cursor.execute(f"""CREATE TABLE {TEST_TABLE_ID} (
                            PersonID int,
                            LastName varchar(255),
                            FirstName varchar(255),
                            Address varchar(255),
                            City varchar(255)
                            )""")
        q = f"""INSERT INTO {TEST_TABLE_ID} (
        PersonID, LastName, FirstName, Address, City
        ) VALUES
        (0, 'Airflow', 'Apache', '1000 N West Street, Suite 1200', 'Wilmington, NC, USA')
        """
        # raise Exception(q)
        self.cursor.execute(q)

    def teardown_method(self):
        self.cursor.execute(f"DROP TABLE {TEST_TABLE_ID}")

    @mock.patch("airflow.providers.google.cloud.transfers.sql_to_gcs.GCSHook")
    def test_execute(self, gcs_hook_mock_class):
        """Test successful run of execute function for JSON"""
        op = MSSQLToGCSOperator(
            task_id=TASK_ID,
            mssql_conn_id=MSSQL_CONN_ID,
            sql=f"SELECT * FROM {TEST_TABLE_ID}",
            bucket=BUCKET,
            filename=JSON_FILENAME,
        )
        gcs_hook_mock = gcs_hook_mock_class.return_value

        def _assert_upload(bucket, obj, tmp_filename, mime_type=None, gzip=False, metadata=None):
            assert bucket == BUCKET
            assert JSON_FILENAME.format(0) == obj
            assert mime_type == "application/json"
            assert gzip == GZIP
            with open(tmp_filename, "rb") as file:
                assert b"".join(NDJSON_LINES) == file.read()

        gcs_hook_mock.upload.side_effect = _assert_upload
        op.execute(context=mock.MagicMock())



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import json
import logging
from unittest import mock

import httplib2
import pytest
from googleapiclient.errors import HttpError
from paramiko.ssh_exception import SSHException

from airflow.exceptions import AirflowException
from airflow.models import Connection
from airflow.providers.google.cloud.hooks.compute_ssh import ComputeEngineSSHHook
from airflow.providers.google.cloud.hooks.os_login import OSLoginHook

TEST_PROJECT_ID = "test-project-id"

TEST_INSTANCE_NAME = "test-instance"
TEST_ZONE = "test-zone-42"
INTERNAL_IP = "192.9.9.9"
EXTERNAL_IP = "192.3.3.3"
TEST_PUB_KEY = "root:NAME AYZ root"
TEST_PUB_KEY2 = "root:NAME MNJ root"
IMPERSONATION_CHAIN = "SERVICE_ACCOUNT"


class TestComputeEngineHookWithPassedProjectId:
    def test_os_login_hook(self, mocker):
        mock_os_login_hook = mocker.patch.object(OSLoginHook, "__init__", return_value=None, spec=OSLoginHook)

        # Default values
        assert ComputeEngineSSHHook()._oslogin_hook
        mock_os_login_hook.assert_called_with(gcp_conn_id="google_cloud_default")

        # Custom conn_id
        assert ComputeEngineSSHHook(gcp_conn_id="gcpssh")._oslogin_hook
        mock_os_login_hook.assert_called_with(gcp_conn_id="gcpssh")

    @pytest.mark.db_test
    @mock.patch("airflow.providers.google.cloud.hooks.compute_ssh.ComputeEngineHook")
    @mock.patch("airflow.providers.google.cloud.hooks.compute_ssh.paramiko")
    @mock.patch("airflow.providers.google.cloud.hooks.compute_ssh._GCloudAuthorizedSSHClient")
    def test_get_conn_default_configuration(self, mock_ssh_client, mock_paramiko, mock_compute_hook, mocker):
        mock_paramiko.SSHException = RuntimeError
        mock_paramiko.RSAKey.generate.return_value.get_name.return_value = "NAME"
        mock_paramiko.RSAKey.generate.return_value.get_base64.return_value = "AYZ"

        mock_compute_hook.return_value.project_id = TEST_PROJECT_ID
        mock_compute_hook.return_value.get_instance_address.return_value = EXTERNAL_IP

        mock_os_login_hook = mocker.patch.object(
            ComputeEngineSSHHook, "_oslogin_hook", spec=OSLoginHook, name="FakeOsLoginHook"
        )
        type(mock_os_login_hook)._get_credentials_email = mock.PropertyMock(
            return_value="test-example@example.org"
        )
        mock_os_login_hook.import_ssh_public_key.return_value.login_profile.posix_accounts = [
            mock.MagicMock(username="test-username")
        ]

        hook = ComputeEngineSSHHook(instance_name=TEST_INSTANCE_NAME, zone=TEST_ZONE)
        result = hook.get_conn()
        assert mock_ssh_client.return_value == result

        mock_paramiko.RSAKey.generate.assert_called_once_with(2048)
        mock_compute_hook.assert_has_calls(
            [
                mock.call(gcp_conn_id="google_cloud_default"),
                mock.call().get_instance_address(
                    project_id=TEST_PROJECT_ID,
                    resource_id=TEST_INSTANCE_NAME,
                    use_internal_ip=False,
                    zone=TEST_ZONE,
                ),
            ]
        )
        mock_os_login_hook.import_ssh_public_key.assert_called_once_with(
            ssh_public_key={"key": "NAME AYZ root", "expiration_time_usec": mock.ANY},
            project_id="test-project-id",
            user="test-example@example.org",
        )
        mock_ssh_client.assert_has_calls(
            [
                mock.call(mock_compute_hook.return_value),
                mock.call().set_missing_host_key_policy(mock_paramiko.AutoAddPolicy.return_value),
                mock.call().connect(
                    hostname=EXTERNAL_IP,
                    look_for_keys=False,
                    pkey=mock_paramiko.RSAKey.generate.return_value,
                    sock=None,
                    username="test-username",
                ),
            ]
        )

    @pytest.mark.db_test
    @pytest.mark.parametrize(
        "exception_type, error_message",
        [(SSHException, r"Error occurred when establishing SSH connection using Paramiko")],
    )
    @mock.patch("airflow.providers.google.cloud.hooks.compute_ssh.ComputeEngineHook")
    @mock.patch("airflow.providers.google.cloud.hooks.compute_ssh.paramiko")
    @mock.patch("airflow.providers.google.cloud.hooks.compute_ssh._GCloudAuthorizedSSHClient")
    @mock.patch("airflow.providers.google.cloud.hooks.compute_ssh.ComputeEngineSSHHook._connect_to_instance")
    def test_get_conn_default_configuration_test_exceptions(
        self,
        mock_connect,
        mock_ssh_client,
        mock_paramiko,
        mock_compute_hook,
        exception_type,
        error_message,
        caplog,
        mocker,
    ):
        mock_paramiko.SSHException = RuntimeError
        mock_paramiko.RSAKey.generate.return_value.get_name.return_value = "NAME"
        mock_paramiko.RSAKey.generate.return_value.get_base64.return_value = "AYZ"

        mock_compute_hook.return_value.project_id = TEST_PROJECT_ID
        mock_compute_hook.return_value.get_instance_address.return_value = EXTERNAL_IP

        mock_os_login_hook = mocker.patch.object(
            ComputeEngineSSHHook, "_oslogin_hook", spec=OSLoginHook, name="FakeOsLoginHook"
        )
        type(mock_os_login_hook)._get_credentials_email = mock.PropertyMock(
            return_value="test-example@example.org"
        )
        mock_os_login_hook.import_ssh_public_key.return_value.login_profile.posix_accounts = [
            mock.MagicMock(username="test-username")
        ]

        hook = ComputeEngineSSHHook(instance_name=TEST_INSTANCE_NAME, zone=TEST_ZONE)
        mock_connect.side_effect = [exception_type, mock_ssh_client]

        with caplog.at_level(logging.INFO):
            hook.get_conn()
        assert error_message in caplog.text
        assert "Failed establish SSH connection" in caplog.text

    @pytest.mark.db_test
    @mock.patch("airflow.providers.google.cloud.hooks.compute_ssh.ComputeEngineHook")
    @mock.patch("airflow.providers.google.cloud.hooks.compute_ssh.OSLoginHook")
    @mock.patch("airflow.providers.google.cloud.hooks.compute_ssh.paramiko")
    @mock.patch("airflow.providers.google.cloud.hooks.compute_ssh._GCloudAuthorizedSSHClient")
    def test_get_conn_authorize_using_instance_metadata(
        self, mock_ssh_client, mock_paramiko, mock_os_login_hook, mock_compute_hook
    ):
        mock_paramiko.SSHException = Exception
        mock_paramiko.RSAKey.generate.return_value.get_name.return_value = "NAME"
        mock_paramiko.RSAKey.generate.return_value.get_base64.return_value = "AYZ"

        mock_compute_hook.return_value.project_id = TEST_PROJECT_ID
        mock_compute_hook.return_value.get_instance_address.return_value = EXTERNAL_IP

        mock_compute_hook.return_value.get_instance_info.return_value = {"metadata": {}}

        hook = ComputeEngineSSHHook(instance_name=TEST_INSTANCE_NAME, zone=TEST_ZONE, use_oslogin=False)
        result = hook.get_conn()
        assert mock_ssh_client.return_value == result

        mock_paramiko.RSAKey.generate.assert_called_once_with(2048)
        mock_compute_hook.assert_has_calls(
            [
                mock.call(gcp_conn_id="google_cloud_default"),
                mock.call().get_instance_address(
                    project_id=TEST_PROJECT_ID,
                    resource_id=TEST_INSTANCE_NAME,
                    use_internal_ip=False,
                    zone=TEST_ZONE,
                ),
                mock.call().get_instance_info(
                    project_id=TEST_PROJECT_ID, resource_id=TEST_INSTANCE_NAME, zone=TEST_ZONE
                ),
                mock.call().set_instance_metadata(
                    metadata={"items": [{"key": "ssh-keys", "value": f"{TEST_PUB_KEY}\n"}]},
                    project_id=TEST_PROJECT_ID,
                    resource_id=TEST_INSTANCE_NAME,
                    zone=TEST_ZONE,
                ),
            ]
        )

        mock_ssh_client.assert_has_calls(
            [
                mock.call(mock_compute_hook.return_value),
                mock.call().set_missing_host_key_policy(mock_paramiko.AutoAddPolicy.return_value),
                mock.call().connect(
                    hostname=EXTERNAL_IP,
                    look_for_keys=False,
                    pkey=mock_paramiko.RSAKey.generate.return_value,
                    sock=None,
                    username="root",
                ),
            ]
        )

        mock_os_login_hook.return_value.import_ssh_public_key.assert_not_called()

    @pytest.mark.db_test
    @pytest.mark.parametrize(
        "exception_type, error_message",
        [
            (
                HttpError(resp=httplib2.Response({"status": 412}), content=b"Error content"),
                r"Error occurred when trying to update instance metadata",
            ),
            (
                AirflowException("412 PRECONDITION FAILED"),
                r"Error occurred when trying to update instance metadata",
            ),
        ],
    )
    @mock.patch("airflow.providers.google.cloud.hooks.compute_ssh.ComputeEngineHook")
    @mock.patch("airflow.providers.google.cloud.hooks.compute_ssh.OSLoginHook")
    @mock.patch("airflow.providers.google.cloud.hooks.compute_ssh.paramiko")
    @mock.patch("airflow.providers.google.cloud.hooks.compute_ssh._GCloudAuthorizedSSHClient")
    def test_get_conn_authorize_using_instance_metadata_test_exception(
        self,
        mock_ssh_client,
        mock_paramiko,
        mock_os_login_hook,
        mock_compute_hook,
        exception_type,
        error_message,
        caplog,
    ):
        mock_paramiko.SSHException = Exception
        mock_paramiko.RSAKey.generate.return_value.get_name.return_value = "NAME"
        mock_paramiko.RSAKey.generate.return_value.get_base64.return_value = "AYZ"

        mock_compute_hook.return_value.project_id = TEST_PROJECT_ID
        mock_compute_hook.return_value.get_instance_address.return_value = EXTERNAL_IP

        mock_compute_hook.return_value.get_instance_info.return_value = {"metadata": {}}
        mock_compute_hook.return_value.set_instance_metadata.side_effect = [exception_type, None]

        hook = ComputeEngineSSHHook(instance_name=TEST_INSTANCE_NAME, zone=TEST_ZONE, use_oslogin=False)
        with caplog.at_level(logging.INFO):
            hook.get_conn()
        assert error_message in caplog.text
        assert "Failed establish SSH connection" in caplog.text

    @pytest.mark.db_test
    @mock.patch("airflow.providers.google.cloud.hooks.compute_ssh.ComputeEngineHook")
    @mock.patch("airflow.providers.google.cloud.hooks.compute_ssh.OSLoginHook")
    @mock.patch("airflow.providers.google.cloud.hooks.compute_ssh.paramiko")
    @mock.patch("airflow.providers.google.cloud.hooks.compute_ssh._GCloudAuthorizedSSHClient")
    def test_get_conn_authorize_using_instance_metadata_append_ssh_keys(
        self, mock_ssh_client, mock_paramiko, mock_os_login_hook, mock_compute_hook
    ):
        del mock_os_login_hook
        mock_paramiko.SSHException = Exception
        mock_paramiko.RSAKey.generate.return_value.get_name.return_value = "NAME"
        mock_paramiko.RSAKey.generate.return_value.get_base64.return_value = "AYZ"

        mock_compute_hook.return_value.project_id = TEST_PROJECT_ID
        mock_compute_hook.return_value.get_instance_address.return_value = EXTERNAL_IP

        mock_compute_hook.return_value.get_instance_info.return_value = {
            "metadata": {"items": [{"key": "ssh-keys", "value": f"{TEST_PUB_KEY2}\n"}]}
        }

        hook = ComputeEngineSSHHook(instance_name=TEST_INSTANCE_NAME, zone=TEST_ZONE, use_oslogin=False)
        result = hook.get_conn()
        assert mock_ssh_client.return_value == result

        mock_compute_hook.return_value.set_instance_metadata.assert_called_once_with(
            metadata={"items": [{"key": "ssh-keys", "value": f"{TEST_PUB_KEY}\n{TEST_PUB_KEY2}\n"}]},
            project_id=TEST_PROJECT_ID,
            resource_id=TEST_INSTANCE_NAME,
            zone=TEST_ZONE,
        )

    @pytest.mark.db_test
    @mock.patch("airflow.providers.google.cloud.hooks.compute_ssh.ComputeEngineHook")
    @mock.patch("airflow.providers.google.cloud.hooks.compute_ssh.OSLoginHook")
    @mock.patch("airflow.providers.google.cloud.hooks.compute_ssh.paramiko")
    @mock.patch("airflow.providers.google.cloud.hooks.compute_ssh._GCloudAuthorizedSSHClient")
    def test_get_conn_private_ip(self, mock_ssh_client, mock_paramiko, mock_os_login_hook, mock_compute_hook):
        del mock_os_login_hook
        mock_paramiko.SSHException = Exception
        mock_paramiko.RSAKey.generate.return_value.get_name.return_value = "NAME"
        mock_paramiko.RSAKey.generate.return_value.get_base64.return_value = "AYZ"

        mock_compute_hook.return_value.project_id = TEST_PROJECT_ID
        mock_compute_hook.return_value.get_instance_address.return_value = INTERNAL_IP

        mock_compute_hook.return_value.get_instance_info.return_value = {"metadata": {}}

        hook = ComputeEngineSSHHook(
            instance_name=TEST_INSTANCE_NAME, zone=TEST_ZONE, use_oslogin=False, use_internal_ip=True
        )
        result = hook.get_conn()
        assert mock_ssh_client.return_value == result

        mock_compute_hook.return_value.get_instance_address.assert_called_once_with(
            project_id=TEST_PROJECT_ID, resource_id=TEST_INSTANCE_NAME, use_internal_ip=True, zone=TEST_ZONE
        )
        mock_ssh_client.return_value.connect.assert_called_once_with(
            hostname=INTERNAL_IP, look_for_keys=mock.ANY, pkey=mock.ANY, sock=mock.ANY, username=mock.ANY
        )

    @pytest.mark.db_test
    @mock.patch("airflow.providers.google.cloud.hooks.compute_ssh.ComputeEngineHook")
    @mock.patch("airflow.providers.google.cloud.hooks.compute_ssh.OSLoginHook")
    @mock.patch("airflow.providers.google.cloud.hooks.compute_ssh.paramiko")
    @mock.patch("airflow.providers.google.cloud.hooks.compute_ssh._GCloudAuthorizedSSHClient")
    def test_get_conn_custom_hostname(
        self, mock_ssh_client, mock_paramiko, mock_os_login_hook, mock_compute_hook
    ):
        del mock_os_login_hook
        mock_paramiko.SSHException = Exception

        hook = ComputeEngineSSHHook(
            instance_name=TEST_INSTANCE_NAME,
            zone=TEST_ZONE,
            use_oslogin=False,
            hostname="custom-hostname",
        )
        result = hook.get_conn()
        assert mock_ssh_client.return_value == result

        mock_compute_hook.return_value.get_instance_address.assert_not_called()
        mock_ssh_client.return_value.connect.assert_called_once_with(
            hostname="custom-hostname",
            look_for_keys=mock.ANY,
            pkey=mock.ANY,
            sock=mock.ANY,
            username=mock.ANY,
        )

    @pytest.mark.db_test
    @mock.patch("airflow.providers.google.cloud.hooks.compute_ssh.ComputeEngineHook")
    @mock.patch("airflow.providers.google.cloud.hooks.compute_ssh.OSLoginHook")
    @mock.patch("airflow.providers.google.cloud.hooks.compute_ssh.paramiko")
    @mock.patch("airflow.providers.google.cloud.hooks.compute_ssh._GCloudAuthorizedSSHClient")
    def test_get_conn_iap_tunnel(self, mock_ssh_client, mock_paramiko, mock_os_login_hook, mock_compute_hook):
        del mock_os_login_hook
        mock_paramiko.SSHException = Exception

        mock_compute_hook.return_value.project_id = TEST_PROJECT_ID

        hook = ComputeEngineSSHHook(
            instance_name=TEST_INSTANCE_NAME, zone=TEST_ZONE, use_oslogin=False, use_iap_tunnel=True
        )
        result = hook.get_conn()
        assert mock_ssh_client.return_value == result

        mock_ssh_client.return_value.connect.assert_called_once_with(
            hostname=mock.ANY,
            look_for_keys=mock.ANY,
            pkey=mock.ANY,
            sock=mock_paramiko.ProxyCommand.return_value,
            username=mock.ANY,
        )
        mock_paramiko.ProxyCommand.assert_called_once_with(
            f"gcloud compute start-iap-tunnel {TEST_INSTANCE_NAME} 22 "
            f"--listen-on-stdin --project={TEST_PROJECT_ID} "
            f"--zone={TEST_ZONE} --verbosity=warning"
        )

    @pytest.mark.db_test
    @mock.patch("airflow.providers.google.cloud.hooks.compute_ssh.ComputeEngineHook")
    @mock.patch("airflow.providers.google.cloud.hooks.compute_ssh.OSLoginHook")
    @mock.patch("airflow.providers.google.cloud.hooks.compute_ssh.paramiko")
    @mock.patch("airflow.providers.google.cloud.hooks.compute_ssh._GCloudAuthorizedSSHClient")
    def test_get_conn_iap_tunnel_with_impersonation_chain(
        self, mock_ssh_client, mock_paramiko, mock_os_login_hook, mock_compute_hook
    ):
        del mock_os_login_hook
        mock_paramiko.SSHException = Exception

        mock_compute_hook.return_value.project_id = TEST_PROJECT_ID

        hook = ComputeEngineSSHHook(
            instance_name=TEST_INSTANCE_NAME,
            zone=TEST_ZONE,
            use_oslogin=False,
            use_iap_tunnel=True,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        result = hook.get_conn()
        assert mock_ssh_client.return_value == result

        mock_ssh_client.return_value.connect.assert_called_once_with(
            hostname=mock.ANY,
            look_for_keys=mock.ANY,
            pkey=mock.ANY,
            sock=mock_paramiko.ProxyCommand.return_value,
            username=mock.ANY,
        )
        mock_paramiko.ProxyCommand.assert_called_once_with(
            f"gcloud compute start-iap-tunnel {TEST_INSTANCE_NAME} 22 "
            f"--listen-on-stdin --project={TEST_PROJECT_ID} "
            f"--zone={TEST_ZONE} --verbosity=warning --impersonate-service-account={IMPERSONATION_CHAIN}"
        )

    @pytest.mark.db_test
    @pytest.mark.parametrize(
        "exception_type, error_message",
        [(SSHException, r"Error occurred when establishing SSH connection using Paramiko")],
    )
    @mock.patch("airflow.providers.google.cloud.hooks.compute_ssh.ComputeEngineHook")
    @mock.patch("airflow.providers.google.cloud.hooks.compute_ssh.OSLoginHook")
    @mock.patch("airflow.providers.google.cloud.hooks.compute_ssh.paramiko")
    @mock.patch("airflow.providers.google.cloud.hooks.compute_ssh._GCloudAuthorizedSSHClient")
    @mock.patch("airflow.providers.google.cloud.hooks.compute_ssh.ComputeEngineSSHHook._connect_to_instance")
    def test_get_conn_iap_tunnel_test_exception(
        self,
        mock_connect,
        mock_ssh_client,
        mock_paramiko,
        mock_os_login_hook,
        mock_compute_hook,
        exception_type,
        error_message,
        caplog,
    ):
        del mock_os_login_hook
        mock_paramiko.SSHException = Exception

        mock_compute_hook.return_value.project_id = TEST_PROJECT_ID

        hook = ComputeEngineSSHHook(
            instance_name=TEST_INSTANCE_NAME, zone=TEST_ZONE, use_oslogin=False, use_iap_tunnel=True
        )
        mock_connect.side_effect = [exception_type, mock_ssh_client]

        with caplog.at_level(logging.INFO):
            hook.get_conn()
        assert error_message in caplog.text
        assert "Failed establish SSH connection" in caplog.text

    @pytest.mark.db_test
    @mock.patch("airflow.providers.google.cloud.hooks.compute_ssh.ComputeEngineHook")
    @mock.patch("airflow.providers.google.cloud.hooks.compute_ssh.OSLoginHook")
    @mock.patch("airflow.providers.google.cloud.hooks.compute_ssh.paramiko")
    @mock.patch("airflow.providers.google.cloud.hooks.compute_ssh._GCloudAuthorizedSSHClient")
    @mock.patch("airflow.providers.google.cloud.hooks.compute_ssh.time.sleep")
    def test_get_conn_retry_on_connection_error(
        self, mock_time, mock_ssh_client, mock_paramiko, mock_os_login_hook, mock_compute_hook
    ):
        del mock_os_login_hook
        del mock_compute_hook

        class CustomException(Exception):
            pass

        mock_paramiko.SSHException = CustomException
        mock_ssh_client.return_value.connect.side_effect = [CustomException, CustomException, True]
        hook = ComputeEngineSSHHook(instance_name=TEST_INSTANCE_NAME, zone=TEST_ZONE)
        hook.get_conn()

        assert mock_ssh_client.return_value.connect.call_count == 3

    def test_read_configuration_from_connection(self):
        conn = Connection(
            conn_type="gcpssh",
            login="conn-user",
            host="conn-host",
            extra=json.dumps(
                {
                    "instance_name": "conn-instance-name",
                    "zone": "zone",
                    "use_internal_ip": True,
                    "use_iap_tunnel": True,
                    "use_oslogin": False,
                    "expire_time": 4242,
                }
            ),
        )
        conn_uri = conn.get_uri()
        with mock.patch.dict("os.environ", AIRFLOW_CONN_GCPSSH=conn_uri):
            hook = ComputeEngineSSHHook(gcp_conn_id="gcpssh")
            hook._load_connection_config()
        assert hook.instance_name == "conn-instance-name"
        assert hook.hostname == "conn-host"
        assert hook.user == "conn-user"
        assert hook.use_internal_ip is True
        assert isinstance(hook.use_internal_ip, bool)
        assert hook.use_iap_tunnel is True
        assert isinstance(hook.use_iap_tunnel, bool)
        assert hook.use_oslogin is False
        assert isinstance(hook.use_oslogin, bool)
        assert hook.expire_time == 4242
        assert isinstance(hook.expire_time, int)

    def test_read_configuration_from_connection_empty_config(self):
        conn = Connection(
            conn_type="gcpssh",
            extra=json.dumps({}),
        )
        conn_uri = conn.get_uri()
        with mock.patch.dict("os.environ", AIRFLOW_CONN_GCPSSH=conn_uri):
            hook = ComputeEngineSSHHook(gcp_conn_id="gcpssh")
            hook._load_connection_config()
        assert hook.instance_name is None
        assert hook.hostname is None
        assert hook.user == "root"
        assert hook.use_internal_ip is False
        assert hook.use_iap_tunnel is False
        assert hook.use_oslogin is False
        assert hook.expire_time == 300
        assert isinstance(hook.expire_time, int)

    @pytest.mark.parametrize(
        "metadata, expected_metadata",
        [
            ({"items": []}, {"items": [{"key": "ssh-keys", "value": "user:pubkey\n"}]}),
            (
                {"items": [{"key": "test", "value": "test"}]},
                {"items": [{"key": "ssh-keys", "value": "user:pubkey\n"}, {"key": "test", "value": "test"}]},
            ),
            (
                {"items": [{"key": "ssh-keys", "value": "test"}, {"key": "test", "value": "test"}]},
                {
                    "items": [
                        {"key": "ssh-keys", "value": "user:pubkey\ntest"},
                        {"key": "test", "value": "test"},
                    ]
                },
            ),
        ],
    )
    @mock.patch("airflow.providers.google.cloud.hooks.compute_ssh.ComputeEngineHook.set_instance_metadata")
    @mock.patch("airflow.providers.google.cloud.hooks.compute_ssh.ComputeEngineHook.get_instance_info")
    def test__authorize_compute_engine_instance_metadata(
        self, mock_get_instance_info, mock_set_instance_metadata, metadata, expected_metadata
    ):
        """Test to ensure the addition metadata is retained"""
        mock_get_instance_info.return_value = {"metadata": metadata}
        conn = Connection(
            conn_type="gcpssh",
            extra=json.dumps({}),
        )
        conn_uri = conn.get_uri()
        with mock.patch.dict("os.environ", AIRFLOW_CONN_GCPSSH=conn_uri):
            hook = ComputeEngineSSHHook(gcp_conn_id="gcpssh")
            hook.user = "user"
            pubkey = "pubkey"
            hook._authorize_compute_engine_instance_metadata(pubkey=pubkey)
            mock_set_instance_metadata.call_args.kwargs["metadata"]["items"].sort(key=lambda x: x["key"])
            expected_metadata["items"].sort(key=lambda x: x["key"])
            assert mock_set_instance_metadata.call_args.kwargs["metadata"] == expected_metadata
